package com.keyware.composeanalysis.controller;

import com.keyware.common.constant.RedisConst;
import com.keyware.composeanalysis.api.ComposeAnalysisApi;
import com.keyware.composeanalysis.constant.enums.AnalysisStatusEnum;
import com.keyware.composeanalysis.entity.AnalysisTask;
import com.keyware.composeanalysis.response.AnalysisResp;
import com.keyware.composeanalysis.service.AnalysisTaskService;
import com.keyware.composeanalysis.util.IpUtil;
import jakarta.annotation.Resource;
import lombok.extern.log4j.Log4j2;
import org.redisson.api.RLock;
import org.redisson.api.RedissonClient;
import org.springframework.web.bind.annotation.RestController;

/**
 * @author liuzongren
 * @date 2024/7/30
 */
@Log4j2
@RestController
public class ComposeAnalysisController implements ComposeAnalysisApi {

    @Resource
    private AnalysisTaskService taskService;

    @Resource
    private RedissonClient redissonClient;

    @Override
    public AnalysisResp startComposeAnalysisTask(String taskId) {
        AnalysisResp result = new AnalysisResp();
        result.setNodeIp(IpUtil.getHostIp());
        RLock lock = redissonClient.getLock(String.format(RedisConst.TASK_LOCK_KEY_PREFIX, taskId));
        try {
            //执行任务前 首先获取当前任务的锁，防止多节点并发分析同一任务
            if (lock.tryLock()) {
                AnalysisTask analysisTask = taskService.getById(taskId);

                //校验任务是否存在
                if (analysisTask == null) {
                    result.setCode(202);
                    result.setResponseMsg("当前任务不存在");
                    return result;
                }

                if (analysisTask.getAnalysisStatus().equals(AnalysisStatusEnum.ANALYSIS_DONE.getCode())) {
                    result.setCode(202);
                    result.setResponseMsg("任务已分析完成,如需重新分析,请点击重新分析");
                    return result;
                }

                //执行成分分析任务
                result.setCode(200);
                taskService.doComposeAnalyze(analysisTask);
                result.setResponseMsg("任务执行成功");
            }
        } catch (Exception e) {
            result.setCode(500);
            result.setResponseMsg("任务执行失败");
            log.error("任务执行失败", e);
            lock.unlock();
        }finally {
            lock.unlock();
        }
        return result;
    }

    @Override
    public void stopComposeAnalysisTask(String taskId) {
        taskService.stopComposeAnalysisTask(taskId);
    }

    @Override
    public AnalysisResp restartComposeAnalysisTask(String taskId) {
        AnalysisResp result = new AnalysisResp();
        result.setNodeIp(IpUtil.getHostIp());
        AnalysisTask analysisTask = taskService.getById(taskId);

        //校验任务是否存在
        if (analysisTask == null) {
            result.setCode(202);
            result.setResponseMsg("当前任务不存在");
            return result;
        }

        Boolean isRestart = taskService.restartComposeAnalysisTask(taskId);

        if (isRestart) {
            result.setCode(200);
            result.setResponseMsg("任务重启成功");
        } else {
            result.setCode(202);
            result.setResponseMsg("任务重启失败");
        }
        return result;
    }

    @Override
    public AnalysisResp recoveryComposeAnalysisTask(String taskId) {
        AnalysisResp result = new AnalysisResp();
        result.setNodeIp(IpUtil.getHostIp());
        AnalysisTask analysisTask = taskService.getById(taskId);

        //校验任务是否存在
        if (analysisTask == null) {
            result.setCode(202);
            result.setResponseMsg("当前任务不存在");
            return result;
        }
        taskService.recoveryComposeAnalysisTask(analysisTask);

        result.setCode(200);
        result.setResponseMsg("任务恢复成功");
        return result;
    }


}
